// This is WXCUBES.CPP. Copyright (C) 2004 by Pawel Pilarczyk.
// It is a program for displaying cubical sets (it uses wxWin and Open GL).
// This is free software. No warranty. Consult 'license.txt' for details.
// Started on June 30, 2004. Last revision: August 7, 2004 [Sep 16, 2004].


// #define _WXCUBES_CPP_

#include "options.h"
#include "textfile.h"
#include "pointset.h"
#include "wxglmain.h"

#include <math.h>

const char *helpinfo = "\
Call this program with one or more names of text files to display cubical \
sets defined in those files. Each cube is of size 1 and is defined by its \
vertex with the minimal coordinates, e.g. (0,1,-1) defines the cube \
[0,1]x[1,2]x[-1,0]. Each set of cubes is displayed in different color \
(there are 6 colors available).";


// --------------------------------------------------
// ------------------ Cubical sets ------------------
// --------------------------------------------------

#ifndef MAXDIM
#define MAXDIM 100
#endif

extern coordinate logo [] [3];

double grid = 2, radius = 1;
double center [3] = {0, 0, 0};

const int maxSets = 32;
int nSets = 0;
pointset p [maxSets];
char *cubefaces [maxSets];

coordinate minimal [3] = {0, 0, 0}, maximal [3] = {0, 0, 0};

static void loadCubes (const char *filename)
{
	if (nSets >= maxSets)
		return;
	
	// read the set of cubes from a file
	std::ifstream in (filename);
	if (!in)
		return;
	ignorecomments (in);
	p [nSets]. dimension (3);
	do
	{
		// find an opening parenthesis if possible
		while (in && (closingparenthesis (in. peek ()) == EOF))
		{
			ignoreline (in);
			ignorecomments (in);
		}
		if (!in)
			break;

		// read the coordinates from the file
		coordinate c [MAXDIM];
		c [0] = c [1] = c [2] = 0;
		int dim = readcoordinates (in, c, MAXDIM);
		if (dim == MAXDIM)
			break;
		ignorecomments (in);
		
		// add the point to the set
		p [nSets]. add (c);

	} while (in);
	in. close ();

	// ignore this set of points if it is empty
	if (!(int) (p [nSets]))
		return;

	// compute the range of the cubes' coordinates
	if ((int) (p [nSets]))
	{
		static bool firsttime = true;
		for (int i = 0; i < 3; ++ i)
		{
			if (firsttime ||
				(minimal [i] > p [nSets]. minimal [i]))
				minimal [i] = p [nSets]. minimal [i];
			if (firsttime ||
				(maximal [i] < p [nSets]. maximal [i]))
				maximal [i] = p [nSets]. maximal [i];
		}
		firsttime = false;
	}

	nSets ++;
	return;
} /* loadCubes */

static void analyzeCubes ()
{
	// determine which faces of the cubes should be displayed
	for (int n = 0; n < nSets; ++ n)
	{
		// determine which faces of the cubes have to be drawn
		cubefaces [n] = new char [1 + (int) (p [n])];
		coordinate c [100];
		for (int i = 0; i < (int) (p [n]); ++ i)
		{
			for (int k = 0; k < 3; ++ k)
				c [k] = p [n] [i] [k];
			char faces = 0;
			char facebit = 0x01;
			for (int j = 0; j < 6; ++ j)
			{
				c [j >> 1] += (j & 1) ? -1 : 1;
				bool found = false;
				for (int k = 0; !found && (k < nSets); ++ k)
					if (p [k]. check (c))
						found = true;
				if (!found)
					faces |= facebit;
				facebit <<= 1;
				c [j >> 1] -= (j & 1) ? -1 : 1;
			}
			cubefaces [n] [i] = faces;
		}
	}

	int i;
	// compute the coordinates of the middle of the set of cubes
	for (i = 0; i < 3; ++ i)
		center [i] = (maximal [i] + minimal [i]) / 2.0;
	// find the diameter of the set in the "maximum" metric
	int diam = 0;
	for (i = 0; i < 3; ++ i)
		if (diam < maximal [i] - minimal [i])
			diam = maximal [i] - minimal [i];
	++ diam;
	// compute the "max"-metric radius of each cube and the grid size
	radius = 1.0 / diam;
	grid = 2 * radius;

	return;
} /* analyzeCubes */


// --------------------------------------------------
// ---------------- Cubes and OpenGL ----------------
// --------------------------------------------------

inline void drawCube (double x, double y, double z, char faces)
// This function draws the relevant faces of the cube whose center is given.
{
	// right
	if (faces & 0x01)
	{
		glNormal3f (1.0, 0.0, 0.0);
		glVertex3f(x + radius, y - radius, z - radius);
		glVertex3f(x + radius, y + radius, z - radius);
		glVertex3f(x + radius, y + radius, z + radius);
		glVertex3f(x + radius, y - radius, z + radius);
	}

	// left
	if (faces & 0x02)
	{
		glNormal3f(-1.0, 0.0, 0.0);
		glVertex3f(x - radius, y - radius, z - radius);
		glVertex3f(x - radius, y - radius, z + radius);
		glVertex3f(x - radius, y + radius, z + radius);
		glVertex3f(x - radius, y + radius, z - radius);
	}

	// back
	if (faces & 0x04)
	{
		glNormal3f (0.0, 1.0, 0.0);
		glVertex3f(x - radius, y + radius, z - radius);
		glVertex3f(x - radius, y + radius, z + radius);
		glVertex3f(x + radius, y + radius, z + radius);
		glVertex3f(x + radius, y + radius, z - radius);
	}

	// front
	if (faces & 0x08)
	{
		glNormal3f (0.0,-1.0, 0.0);
		glVertex3f(x - radius, y - radius, z - radius);
		glVertex3f(x + radius, y - radius, z - radius);
		glVertex3f(x + radius, y - radius, z + radius);
		glVertex3f(x - radius, y - radius, z + radius);
	}

	// top
	if (faces & 0x10)
	{
		glNormal3f (0.0, 0.0, 1.0);
		glVertex3f(x - radius, y - radius, z + radius);
		glVertex3f(x + radius, y - radius, z + radius);
		glVertex3f(x + radius, y + radius, z + radius);
		glVertex3f(x - radius, y + radius, z + radius);
	}

	// bottom
	if (faces & 0x20)
	{
	 	glNormal3f (0.0, 0.0,-1.0);
		glVertex3f(x - radius, y - radius, z - radius);
		glVertex3f(x - radius, y + radius, z - radius);
		glVertex3f(x + radius, y + radius, z - radius);
		glVertex3f(x + radius, y - radius, z - radius);
	}

	return;
} /* drawCube */

inline void drawCube (coordinate *c, char faces)
// This is a wrapper function for drawing a cube.
{
	drawCube ((c [0] - center [0]) * grid,
		(c [1] - center [1]) * grid,
		(c [2] - center [2]) * grid, faces);
	return; 
} /* drawCube */


// --------------------------------------------------
// ----------------- OpenGL drawing -----------------
// --------------------------------------------------

void initPicture (int argc, char *argv [])
// This function is called at canvas initialization.
// The function "resizePicture" is called immediately afterwards.
{
	// load the set of cubes to be displayed and analyze them
	for (int i = 1; i < argc; ++ i)
		loadCubes (argv [i]);

	// if nothing was read, create a default set of cubes
	if (nSets == 0)
	{
		int i = 0;
		do
		{
			p [nSets]. dimension (3);
			while (logo [i] [0] || logo [i] [1] || logo [i] [2])
				p [nSets]. add (logo [i ++]);
			nSets ++;
			i ++;
		} while (logo [i] [0] || logo [i] [1] || logo [i] [2]);
		maximal [0] = 101;
		maximal [1] = 55;
		maximal [2] = 30;
	}

	// analyze the sets of cubes
	analyzeCubes ();

	// enable OpenGL lighting
	glEnable (GL_LIGHTING);

	// light number 0
	static GLfloat Ambient0 [4] = { 0.0, 0.2, 0.3, 1.0};
	static GLfloat Diffuse0 [4] = { 0.3, 0.3, 0.5, 1.0};
	static GLfloat Specular0 [4] = { 0.3, 0.3, 1.0, 1.0};
	static GLfloat Position0 [4] = { 0.0, 1.0, 2.0, 0.0};
	glLightfv (GL_LIGHT0, GL_AMBIENT, Ambient0);
	glLightfv (GL_LIGHT0, GL_DIFFUSE, Diffuse0);
	glLightfv (GL_LIGHT0, GL_SPECULAR, Specular0);
	glLightfv (GL_LIGHT0, GL_POSITION, Position0);
	glEnable (GL_LIGHT0);

	// light number 1
	static GLfloat Ambient1 [4] = { 0.3, 0.1, 0.0, 1.0};
	static GLfloat Diffuse1 [4] = { 0.5, 0.3, 0.3, 1.0};
	static GLfloat Specular1 [4] = { 1.0, 0.3, 0.3, 1.0};
	static GLfloat Position1 [4] = { 2.0, 1.0, 0.0, 0.0};
	glLightfv (GL_LIGHT1, GL_AMBIENT, Ambient1);
	glLightfv (GL_LIGHT1, GL_DIFFUSE, Diffuse1);
	glLightfv (GL_LIGHT1, GL_SPECULAR, Specular1);
	glLightfv (GL_LIGHT1, GL_POSITION, Position1);
	glEnable (GL_LIGHT1);

	// light number 7 (used only for the wireframe display)
	static GLfloat Ambient7 [4] = { 1.0, 1.0, 1.0, 1.0};
	glLightfv (GL_LIGHT7, GL_AMBIENT, Ambient7);
	glDisable (GL_LIGHT7);

	// enable the depth test
	glEnable (GL_DEPTH_TEST);
	
	// don't draw interior faces
	glCullFace (GL_BACK);

	// define the fog in the scene
	glFogi (GL_FOG_MODE, GL_EXP2);
	glFogf (GL_FOG_DENSITY, 0.4);
	GLfloat fogColor [] = {1.0, 1.0, 1.0, 1.0};
	glFogfv (GL_FOG_COLOR, fogColor);

	return;
} /* initPicture */

void resizePicture (int width, int height)
// This function is called on canvas resize and when the magnification
// exponent, the translation vector or the projection mode have changed.
{
	// initialize the projection matrices
	glMatrixMode (GL_PROJECTION);
	glLoadIdentity ();

	// calculate the correct perspective projection
	double w, h;
	if (!width)
		width = 1;
	if (!height)
		height = 1;
	if (width < height)
	{
		w = exp (-magnification - (perspective ? 0.45 : -0.2));
		h = w * height / width;
	}
	else
	{
		h = exp (-magnification - (perspective ? 0.45 : -0.2));
		w = h * width / height;
	}

	// define the perspective projection matrix
	if (perspective)
		glFrustum (-w - translation [0], w - translation [0],
			-h - translation [1], h - translation [1], 2.0, 6.0);
	else
		glOrtho (-w - translation [0], w - translation [0],
			-h - translation [1], h - translation [1], 2.0, 6.0);
	glTranslatef (0.0f, 0.0f, -4.75f);

	// prepare and reset the model view matrix
	glMatrixMode (GL_MODELVIEW);
	glLoadIdentity();

	return;
} /* resizePicture */

void drawPicture ()
// This function is called on every paint event to draw the actual picture.
{
	// clear the picture buffers
	glClearColor (bgcolor [0], bgcolor [1], bgcolor [2], bgcolor [3]);
	glClearDepth(1.0f);				// Depth Buffer Setup
	glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	// load the rotation matrix to the model view
	glLoadIdentity ();
	glTranslatef (0.0f, 0.0f, 0.75f);
	glMultMatrixf (rotation);

	if (wireframe)
	{
		glPolygonMode (GL_FRONT_AND_BACK, GL_LINE);
		// draw all the interior faces
		glDisable (GL_CULL_FACE);
		// turn on the full-brightness ambient light
		glEnable (GL_LIGHT7);
		// turn off all the other lights
		glDisable (GL_LIGHT0);
		glDisable (GL_LIGHT1);
	}
	else
	{
		glPolygonMode (GL_FRONT_AND_BACK, GL_FILL);
		// don't draw the interior faces
		glEnable (GL_CULL_FACE);
		// turn off the wireframe light
		glDisable (GL_LIGHT7);
		// turn on all the other lights
		glEnable (GL_LIGHT0);
		glEnable (GL_LIGHT1);
	}

	if (fog)
		glEnable (GL_FOG);
	else
		glDisable (GL_FOG);
	
	for (int i = 0; i < nSets; ++ i)
	{
		if (!(int) p [i])
			continue;

		// define the material properties
		const int nMaterials = 6;
		static GLfloat Material [nMaterials] [4] =
		{
			{0.5, 0.5, 0.9, 1.0},
			{0.9, 0.5, 0.5, 1.0},
			{0.8, 0.8, 0.4, 1.0},
			{0.4, 0.8, 0.8, 1.0},
			{0.8, 0.4, 0.8, 1.0},
			{0.5, 0.8, 0.5, 1.0},
		};
		glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE,
			Material [i % nMaterials]);
		glColorMaterial (GL_FRONT_AND_BACK, GL_SPECULAR);
		glMateriali (GL_FRONT_AND_BACK, GL_SHININESS, 128);

		// enable the material color analysis
		glEnable (GL_COLOR_MATERIAL);

		// draw all the relevant faces of the cubes
		glBegin (GL_QUADS);
		for (int j = 0; j < (int) (p [i]); ++ j)
			if (cubefaces [i] [j])
				drawCube (p [i] [j], wireframe ? 0xFF :
					cubefaces [i] [j]);
		glEnd();
	}

	return;
} /* drawPicture */

void deletePicture (void)
{
	for (int i = 0; i < nSets; ++ i)
		if (cubefaces [i])
		{
			delete [] (cubefaces [i]);
			cubefaces [i] = NULL;
		}

	return;
} /* deletePicture */

coordinate logo [] [3] =
{
	{68,0,30}, {85,0,30}, {68,1,30}, {86,1,30}, {2,2,30},
	{3,2,30}, {8,2,30}, {9,2,30}, {10,2,30}, {11,2,30},
	{14,2,30}, {16,2,30}, {17,2,30}, {22,2,30}, {23,2,30},
	{24,2,30}, {29,2,30}, {30,2,30}, {31,2,30}, {38,2,30},
	{39,2,30}, {42,2,30}, {43,2,30}, {44,2,30}, {52,2,30},
	{53,2,30}, {54,2,30}, {55,2,30}, {58,2,30}, {62,2,30},
	{63,2,30}, {64,2,30}, {68,2,30}, {70,2,30}, {71,2,30},
	{75,2,30}, {79,2,30}, {80,2,30}, {82,2,30}, {86,2,30},
	{90,2,30}, {1,3,30}, {4,3,30}, {7,3,30}, {11,3,30},
	{14,3,30}, {15,3,30}, {18,3,30}, {21,3,30}, {25,3,30},
	{28,3,30}, {32,3,30}, {38,3,30}, {41,3,30}, {45,3,30},
	{51,3,30}, {55,3,30}, {58,3,30}, {61,3,30}, {65,3,30},
	{68,3,30}, {69,3,30}, {72,3,30}, {75,3,30}, {78,3,30},
	{81,3,30}, {82,3,30}, {86,3,30}, {1,4,30}, {7,4,30},
	{11,4,30}, {14,4,30}, {18,4,30}, {21,4,30}, {32,4,30},
	{38,4,30}, {41,4,30}, {45,4,30}, {51,4,30}, {55,4,30},
	{58,4,30}, {65,4,30}, {68,4,30}, {72,4,30}, {75,4,30},
	{78,4,30}, {82,4,30}, {85,4,30}, {87,4,30}, {1,5,30},
	{7,5,30}, {11,5,30}, {14,5,30}, {18,5,30}, {21,5,30},
	{22,5,30}, {23,5,30}, {24,5,30}, {25,5,30}, {29,5,30},
	{30,5,30}, {31,5,30}, {38,5,30}, {41,5,30}, {45,5,30},
	{51,5,30}, {55,5,30}, {58,5,30}, {62,5,30}, {63,5,30},
	{64,5,30}, {68,5,30}, {72,5,30}, {75,5,30}, {79,5,30},
	{80,5,30}, {81,5,30}, {82,5,30}, {85,5,30}, {87,5,30},
	{1,6,30}, {7,6,30}, {11,6,30}, {14,6,30}, {18,6,30},
	{21,6,30}, {25,6,30}, {28,6,30}, {38,6,30}, {41,6,30},
	{45,6,30}, {51,6,30}, {55,6,30}, {58,6,30}, {61,6,30},
	{68,6,30}, {72,6,30}, {75,6,30}, {82,6,30}, {85,6,30},
	{87,6,30}, {1,7,30}, {4,7,30}, {7,7,30}, {11,7,30},
	{14,7,30}, {15,7,30}, {18,7,30}, {21,7,30}, {25,7,30},
	{28,7,30}, {32,7,30}, {38,7,30}, {41,7,30}, {45,7,30},
	{51,7,30}, {54,7,30}, {55,7,30}, {58,7,30}, {61,7,30},
	{65,7,30}, {68,7,30}, {69,7,30}, {72,7,30}, {75,7,30},
	{78,7,30}, {82,7,30}, {84,7,30}, {88,7,30}, {2,8,30},
	{3,8,30}, {7,8,30}, {11,8,30}, {14,8,30}, {16,8,30},
	{17,8,30}, {22,8,30}, {23,8,30}, {24,8,30}, {29,8,30},
	{30,8,30}, {31,8,30}, {37,8,30}, {38,8,30}, {39,8,30},
	{42,8,30}, {43,8,30}, {44,8,30}, {52,8,30}, {53,8,30},
	{55,8,30}, {58,8,30}, {62,8,30}, {63,8,30}, {64,8,30},
	{68,8,30}, {70,8,30}, {71,8,30}, {75,8,30}, {79,8,30},
	{80,8,30}, {81,8,30}, {84,8,30}, {88,8,30}, {14,9,30},
	{38,9,30}, {55,9,30}, {75,9,30}, {14,10,30}, {38,10,30},
	{55,10,30}, {58,10,30}, {75,10,30}, {0,0,0},
	
	{51,15,20}, {52,15,20}, {53,15,20}, {54,15,20}, {55,16,20},
	{2,17,20}, {3,17,20}, {8,17,20}, {9,17,20}, {10,17,20},
	{14,17,20}, {18,17,20}, {21,17,20}, {22,17,20}, {25,17,20},
	{26,17,20}, {28,17,20}, {31,17,20}, {34,17,20}, {38,17,20},
	{41,17,20}, {44,17,20}, {48,17,20}, {52,17,20}, {53,17,20},
	{55,17,20}, {62,17,20}, {63,17,20}, {64,17,20}, {69,17,20},
	{70,17,20}, {71,17,20}, {75,17,20}, {76,17,20}, {79,17,20},
	{80,17,20}, {81,17,20}, {89,17,20}, {90,17,20}, {91,17,20},
	{95,17,20}, {1,18,20}, {4,18,20}, {7,18,20}, {11,18,20},
	{14,18,20}, {18,18,20}, {21,18,20}, {24,18,20}, {27,18,20},
	{28,18,20}, {31,18,20}, {34,18,20}, {38,18,20}, {41,18,20},
	{44,18,20}, {48,18,20}, {51,18,20}, {54,18,20}, {55,18,20},
	{61,18,20}, {65,18,20}, {68,18,20}, {72,18,20}, {75,18,20},
	{78,18,20}, {82,18,20}, {88,18,20}, {92,18,20}, {95,18,20},
	{1,19,20}, {7,19,20}, {11,19,20}, {14,19,20}, {18,19,20},
	{21,19,20}, {24,19,20}, {28,19,20}, {31,19,20}, {34,19,20},
	{38,19,20}, {41,19,20}, {44,19,20}, {48,19,20}, {51,19,20},
	{55,19,20}, {65,19,20}, {68,19,20}, {75,19,20}, {82,19,20},
	{88,19,20}, {92,19,20}, {95,19,20}, {1,20,20}, {7,20,20},
	{11,20,20}, {14,20,20}, {18,20,20}, {21,20,20}, {25,20,20},
	{26,20,20}, {27,20,20}, {28,20,20}, {31,20,20}, {34,20,20},
	{38,20,20}, {41,20,20}, {44,20,20}, {48,20,20}, {51,20,20},
	{55,20,20}, {62,20,20}, {63,20,20}, {64,20,20}, {68,20,20},
	{69,20,20}, {70,20,20}, {71,20,20}, {72,20,20}, {75,20,20},
	{79,20,20}, {80,20,20}, {81,20,20}, {88,20,20}, {92,20,20},
	{95,20,20}, {1,21,20}, {7,21,20}, {11,21,20}, {14,21,20},
	{18,21,20}, {21,21,20}, {28,21,20}, {31,21,20}, {34,21,20},
	{38,21,20}, {41,21,20}, {44,21,20}, {48,21,20}, {51,21,20},
	{55,21,20}, {61,21,20}, {68,21,20}, {72,21,20}, {75,21,20},
	{78,21,20}, {88,21,20}, {92,21,20}, {95,21,20}, {1,22,20},
	{4,22,20}, {7,22,20}, {11,22,20}, {14,22,20}, {15,22,20},
	{18,22,20}, {21,22,20}, {24,22,20}, {28,22,20}, {31,22,20},
	{34,22,20}, {35,22,20}, {38,22,20}, {41,22,20}, {44,22,20},
	{45,22,20}, {48,22,20}, {51,22,20}, {54,22,20}, {55,22,20},
	{61,22,20}, {65,22,20}, {68,22,20}, {72,22,20}, {75,22,20},
	{78,22,20}, {82,22,20}, {88,22,20}, {92,22,20}, {95,22,20},
	{2,23,20}, {3,23,20}, {8,23,20}, {9,23,20}, {10,23,20},
	{14,23,20}, {16,23,20}, {17,23,20}, {20,23,20}, {21,23,20},
	{22,23,20}, {25,23,20}, {26,23,20}, {27,23,20}, {31,23,20},
	{34,23,20}, {36,23,20}, {37,23,20}, {41,23,20}, {44,23,20},
	{46,23,20}, {47,23,20}, {52,23,20}, {53,23,20}, {55,23,20},
	{62,23,20}, {63,23,20}, {64,23,20}, {69,23,20}, {70,23,20},
	{71,23,20}, {74,23,20}, {75,23,20}, {76,23,20}, {79,23,20},
	{80,23,20}, {81,23,20}, {89,23,20}, {90,23,20}, {91,23,20},
	{94,23,20}, {95,23,20}, {96,23,20}, {21,24,20}, {75,24,20},
	{95,24,20}, {21,25,20}, {31,25,20}, {41,25,20}, {75,25,20},
	{96,25,20}, {97,25,20}, {0,0,0},
	
	{2,32,10}, {6,32,10}, {10,32,10}, {13,32,10}, {14,32,10},
	{16,32,10}, {20,32,10}, {26,32,10}, {30,32,10}, {34,32,10},
	{35,32,10}, {37,32,10}, {40,32,10}, {44,32,10}, {48,32,10},
	{52,32,10}, {53,32,10}, {54,32,10}, {59,32,10}, {60,32,10},
	{61,32,10}, {69,32,10}, {70,32,10}, {71,32,10}, {75,32,10},
	{81,32,10}, {84,32,10}, {87,32,10}, {91,32,10}, {92,32,10},
	{93,32,10}, {98,32,10}, {99,32,10}, {100,32,10}, {2,33,10},
	{6,33,10}, {10,33,10}, {13,33,10}, {16,33,10}, {20,33,10},
	{26,33,10}, {30,33,10}, {33,33,10}, {36,33,10}, {37,33,10},
	{40,33,10}, {44,33,10}, {48,33,10}, {51,33,10}, {55,33,10},
	{58,33,10}, {62,33,10}, {68,33,10}, {72,33,10}, {75,33,10},
	{81,33,10}, {84,33,10}, {87,33,10}, {90,33,10}, {94,33,10},
	{97,33,10}, {101,33,10}, {1,34,10}, {3,34,10}, {5,34,10},
	{7,34,10}, {10,34,10}, {13,34,10}, {16,34,10}, {20,34,10},
	{26,34,10}, {30,34,10}, {33,34,10}, {37,34,10}, {40,34,10},
	{44,34,10}, {48,34,10}, {51,34,10}, {62,34,10}, {68,34,10},
	{72,34,10}, {75,34,10}, {81,34,10}, {84,34,10}, {87,34,10},
	{90,34,10}, {101,34,10}, {1,35,10}, {3,35,10}, {5,35,10},
	{7,35,10}, {10,35,10}, {13,35,10}, {16,35,10}, {20,35,10},
	{26,35,10}, {30,35,10}, {34,35,10}, {35,35,10}, {36,35,10},
	{37,35,10}, {40,35,10}, {44,35,10}, {48,35,10}, {51,35,10},
	{52,35,10}, {53,35,10}, {54,35,10}, {55,35,10}, {59,35,10},
	{60,35,10}, {61,35,10}, {68,35,10}, {72,35,10}, {75,35,10},
	{81,35,10}, {84,35,10}, {87,35,10}, {90,35,10}, {91,35,10},
	{92,35,10}, {93,35,10}, {94,35,10}, {98,35,10}, {99,35,10},
	{100,35,10}, {1,36,10}, {3,36,10}, {5,36,10}, {7,36,10},
	{10,36,10}, {13,36,10}, {16,36,10}, {20,36,10}, {26,36,10},
	{30,36,10}, {37,36,10}, {40,36,10}, {44,36,10}, {48,36,10},
	{51,36,10}, {55,36,10}, {58,36,10}, {68,36,10}, {72,36,10},
	{75,36,10}, {81,36,10}, {84,36,10}, {87,36,10}, {90,36,10},
	{94,36,10}, {97,36,10}, {0,37,10}, {4,37,10}, {8,37,10},
	{10,37,10}, {13,37,10}, {16,37,10}, {17,37,10}, {20,37,10},
	{26,37,10}, {27,37,10}, {30,37,10}, {33,37,10}, {37,37,10},
	{40,37,10}, {41,37,10}, {44,37,10}, {45,37,10}, {48,37,10},
	{51,37,10}, {55,37,10}, {58,37,10}, {62,37,10}, {68,37,10},
	{72,37,10}, {75,37,10}, {81,37,10}, {84,37,10}, {87,37,10},
	{90,37,10}, {94,37,10}, {97,37,10}, {101,37,10}, {0,38,10},
	{4,38,10}, {8,38,10}, {10,38,10}, {12,38,10}, {13,38,10},
	{14,38,10}, {16,38,10}, {18,38,10}, {19,38,10}, {26,38,10},
	{28,38,10}, {29,38,10}, {34,38,10}, {35,38,10}, {36,38,10},
	{40,38,10}, {42,38,10}, {43,38,10}, {46,38,10}, {47,38,10},
	{52,38,10}, {53,38,10}, {54,38,10}, {59,38,10}, {60,38,10},
	{61,38,10}, {69,38,10}, {70,38,10}, {71,38,10}, {74,38,10},
	{75,38,10}, {76,38,10}, {80,38,10}, {81,38,10}, {82,38,10},
	{84,38,10}, {87,38,10}, {91,38,10}, {92,38,10}, {93,38,10},
	{98,38,10}, {99,38,10}, {100,38,10}, {13,39,10}, {16,39,10},
	{75,39,10}, {81,39,10}, {87,39,10}, {10,40,10}, {13,40,10},
	{16,40,10}, {76,40,10}, {77,40,10}, {82,40,10}, {83,40,10},
	{84,40,10}, {87,40,10}, {0,0,0},
	
	{49,45,0}, {67,45,0}, {68,45,0}, {69,45,0}, {70,45,0},
	{49,46,0}, {71,46,0}, {3,47,0}, {4,47,0}, {5,47,0},
	{11,47,0}, {12,47,0}, {14,47,0}, {17,47,0}, {20,47,0},
	{26,47,0}, {27,47,0}, {29,47,0}, {33,47,0}, {36,47,0},
	{40,47,0}, {41,47,0}, {42,47,0}, {49,47,0}, {51,47,0},
	{52,47,0}, {56,47,0}, {61,47,0}, {62,47,0}, {63,47,0},
	{68,47,0}, {69,47,0}, {71,47,0}, {74,47,0}, {79,47,0},
	{80,47,0}, {82,47,0}, {85,47,0}, {89,47,0}, {93,47,0},
	{2,48,0}, {6,48,0}, {10,48,0}, {13,48,0}, {14,48,0},
	{17,48,0}, {20,48,0}, {26,48,0}, {29,48,0}, {33,48,0},
	{36,48,0}, {39,48,0}, {43,48,0}, {49,48,0}, {50,48,0},
	{53,48,0}, {56,48,0}, {60,48,0}, {64,48,0}, {67,48,0},
	{70,48,0}, {71,48,0}, {74,48,0}, {78,48,0}, {81,48,0},
	{82,48,0}, {85,48,0}, {89,48,0}, {93,48,0}, {1,49,0},
	{7,49,0}, {10,49,0}, {14,49,0}, {17,49,0}, {20,49,0},
	{26,49,0}, {29,49,0}, {33,49,0}, {36,49,0}, {43,49,0},
	{49,49,0}, {53,49,0}, {56,49,0}, {60,49,0}, {64,49,0},
	{67,49,0}, {71,49,0}, {74,49,0}, {78,49,0}, {82,49,0},
	{85,49,0}, {89,49,0}, {93,49,0}, {1,50,0}, {11,50,0},
	{12,50,0}, {13,50,0}, {14,50,0}, {17,50,0}, {20,50,0},
	{26,50,0}, {29,50,0}, {33,50,0}, {36,50,0}, {40,50,0},
	{41,50,0}, {42,50,0}, {49,50,0}, {53,50,0}, {56,50,0},
	{60,50,0}, {64,50,0}, {67,50,0}, {71,50,0}, {74,50,0},
	{79,50,0}, {80,50,0}, {81,50,0}, {82,50,0}, {85,50,0},
	{89,50,0}, {93,50,0}, {1,51,0}, {14,51,0}, {17,51,0},
	{20,51,0}, {26,51,0}, {29,51,0}, {33,51,0}, {36,51,0},
	{39,51,0}, {49,51,0}, {53,51,0}, {56,51,0}, {60,51,0},
	{64,51,0}, {67,51,0}, {71,51,0}, {74,51,0}, {82,51,0},
	{85,51,0}, {89,51,0}, {93,51,0}, {1,52,0}, {10,52,0},
	{14,52,0}, {17,52,0}, {20,52,0}, {26,52,0}, {29,52,0},
	{30,52,0}, {33,52,0}, {36,52,0}, {39,52,0}, {43,52,0},
	{49,52,0}, {50,52,0}, {53,52,0}, {56,52,0}, {57,52,0},
	{60,52,0}, {64,52,0}, {67,52,0}, {70,52,0}, {71,52,0},
	{74,52,0}, {75,52,0}, {78,52,0}, {82,52,0}, {85,52,0},
	{86,52,0}, {89,52,0}, {90,52,0}, {93,52,0}, {1,53,0},
	{7,53,0}, {11,53,0}, {12,53,0}, {13,53,0}, {17,53,0},
	{20,53,0}, {25,53,0}, {26,53,0}, {27,53,0}, {29,53,0},
	{31,53,0}, {32,53,0}, {36,53,0}, {40,53,0}, {41,53,0},
	{42,53,0}, {49,53,0}, {51,53,0}, {52,53,0}, {56,53,0},
	{58,53,0}, {61,53,0}, {62,53,0}, {63,53,0}, {68,53,0},
	{69,53,0}, {71,53,0}, {74,53,0}, {76,53,0}, {79,53,0},
	{80,53,0}, {81,53,0}, {85,53,0}, {87,53,0}, {88,53,0},
	{91,53,0}, {92,53,0}, {2,54,0}, {6,54,0}, {17,54,0},
	{20,54,0}, {26,54,0}, {29,54,0}, {3,55,0}, {4,55,0},
	{5,55,0}, {17,55,0}, {20,55,0}, {26,55,0}, {29,55,0},
	{36,55,0}, {0,0,0}, {0,0,0},
};
